const flat = require('./lib/flat');
const crypto = require('./lib/crypto');
const debug = require('debug')('kkk:sess');

const SESS_VERIFY_KEY = '__sess_verify__';

let opts = {
  name: 'kkk:sess', // session cookie name
  secret: 'kkk:sess:secret', // session crypto key
  crypto: false,
  verify: 'strict', // verify mode, strict will check user-agent
  cookie: {
    // https://github.com/pillarjs/cookies
    signed: true, // sign the cookie value
    overwrite: true, // overwrite the same name (false by default)
  },
};

function get_verify_info(ctx = {}) {
  let keys = opts.keys.join(',');
  let userAgent = (ctx.headers || {})['user-agent'] || 'unknown';
  return crypto.md5(`${keys}:${userAgent}`);
}

function clear_cookie(ctx, name) {
  cooset(ctx, name, '', { expires: new Date(0) });
}

function reset(ctx) {
  clear_cookie(ctx, opts.name);
}

function cooget(ctx, name) {
  return decodeURIComponent(ctx.cookies.get(name, { ...opts.cookie }) || '');
}

function cooset(ctx, name, value, opt) {
  return ctx.cookies.set(name, encodeURIComponent(value), { ...opts.cookie, ...opt });
}

function set(ctx, data) {
  // data null
  if (!Object.keys(data).length) {
    return;
  };

  // init verify info
  if (opts.verify === 'strict') {
    data[SESS_VERIFY_KEY] = get_verify_info(ctx);
  }

  // create new session
  let newsess = '';
  if (opts.crypto) {
    newsess = crypto.encrypt(flat.stringify(opts.keys, data), opts.secret);
  } else {
    newsess = flat.stringify(opts.keys, data);
  }

  // check update
  newsess !== cooget(ctx, opts.name) &&
    cooset(ctx, opts.name, newsess, opts.cookie);

  // clear verify info
  delete data[SESS_VERIFY_KEY];
}

function get(ctx) {
  let result = {};
  try {
    let data = cooget(ctx, opts.name);
    if (data) {
      if (opts.crypto) {
        data = crypto.decrypt(data, opts.secret);
      }
      result = flat.parse(opts.keys, data);
      if (opts.verify === 'strict') {
        result[SESS_VERIFY_KEY] === get_verify_info(ctx) || (result = {});
      }
    }
  } catch (error) {
    debug('get session error:', error);
  } finally {
    // clear verify info
    delete result[SESS_VERIFY_KEY];
  }
  return result;
}

function use(options) {
  opts = Object.assign(opts, options);
  opts.keys = Array.isArray(opts.keys) ? opts.keys : [];
  if (!opts.crypto) {
    opts.verify = '';
  }
  debug('options:', opts);
  if (opts.verify === 'strict') {
    opts.keys.unshift(SESS_VERIFY_KEY);
  }
  return async function (ctx, next) {
    ctx.session = get(ctx);
    ctx.session_reset = reset;
    try {
      await next();
    } catch (ex) {
      throw ex;
    } finally {
      set(ctx, ctx.session);
    }
  };
}

module.exports = {
  get,
  set,
  use,
  reset,
}
